# coding=utf-8
"""Diagnostic to compute regional averages"""
import iris
import iris.util
import iris.coords
import iris.analysis
import iris.exceptions
from iris.coord_categorisation import add_categorised_coord
from iris.cube import Cube, CubeList

import numpy as np
import numba

import netCDF4

from earthdiagnostics.box import Box
from earthdiagnostics.constants import Basins
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticOption, DiagnosticIntOption, DiagnosticDomainOption, \
    DiagnosticBoolOption, DiagnosticBasinListOption, DiagnosticVariableOption
from earthdiagnostics.modelingrealm import ModelingRealms
from earthdiagnostics.utils import Utils, TempFile

import diagonals.zonmean as zonmean
from diagonals.mesh_helpers.nemo import Nemo


class ZonalMean(Diagnostic):
    """
    Computes the zonal mean value of the field (weighted).

    :param data_manager: data management object
    :type data_manager: DataManager
    :param startdate: startdate
    :type startdate: str
    :param member: member number
    :type member: int
    :param chunk: chunk's number
    :type chunk: int
    :param variable: variable to average
    :type variable: str
    :param box: box used to restrict the vertical mean
    :type box: Box
    """

    alias = 'zonmean'
    "Diagnostic alias for the configuration file"

    def __init__(self, data_manager, startdate, member, chunk, domain, variable, basins, grid_point):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.domain = domain
        self.variable = variable
        self.basins = basins
        self.grid_point = grid_point

        self.declared = {}

        self.lat_name = 'lat'
        self.lon_name = 'lon'

    def __eq__(self, other):
        if self._different_type(other):
            return False
        return self.startdate == other.startdate and self.member == other.member and self.chunk == other.chunk and \
            self.variable == other.variable and self.grid_point == other.grid_point and self.basin == other.basin

    def __str__(self):
        return 'Zonal mean Startdate: {0.startdate} Member: {0.member} Chunk: {0.chunk} Variable: {0.variable} ' \
               'Grid point: {0.grid_point}'.format(self)

    def __hash__(self):
        return hash(str(self))

    @classmethod
    def generate_jobs(cls, diags, options):
        """
        Create a job for each chunk to compute the diagnostic

        :param diags: Diagnostics manager class
        :type diags: Diags
        :param options: variable, minimum depth (level), maximum depth (level)
        :type options: list[str]
        :return:
        """
        options_available = (
            DiagnosticDomainOption(),
            DiagnosticVariableOption(diags.data_manager.config.var_manager),
            DiagnosticOption('grid_point', 'T'),
            DiagnosticBasinListOption('basins', Basins().Global),
        )
        options = cls.process_options(options, options_available)

        job_list = list()
        for startdate, member, chunk in diags.config.experiment.get_chunk_list():
            job = ZonalMean(diags.data_manager, startdate, member, chunk,
                            options['domain'], options['variable'], options['basins'],
                            options['grid_point'].lower())
            job_list.append(job)

        return job_list

    def request_data(self):
        """Request data required by the diagnostic"""
        self.variable_file = self.request_chunk(
            self.domain, self.variable, self.startdate, self.member, self.chunk
        )

    def declare_data_generated(self):
        """Declare data to be generated by the diagnostic"""
        self.zonal_mean = self.declare_chunk(
            ModelingRealms.ocean, self.variable + 'zonal',
            self.startdate, self.member, self.chunk, region=self.basins
        )

    def compute(self):
        """Run the diagnostic"""
        self._fix_file_metadata()
        mesh = Nemo('mesh_hgr.nc', 'mask_regions.nc')
        areacello = mesh.get_areacello(cell_point=self.grid_point)
        lats = mesh.get_mesh_var('latitude', dtype=np.float32)
        data = self._load_data()
        masks = {}
        self.basins.sort()
        for basin in self.basins:
            masks[basin] = Utils.get_mask(basin)
        area_basin = zonmean.get_basin_area(areacello, masks)
        zonalmean = zonmean.compute_zonmean(data.data, lats, area_basin)
        self._save_result(zonalmean, data)

    def _try_load_cube(self, number):
        try:
            cube = iris.load_cube('mesh_hgr.nc', 'e{0}{1}'.format(number, self.grid_point))
        except iris.exceptions.ConstraintMismatchError:
            cube = iris.load_cube('mesh_hgr.nc', 'e{0}{1}_0'.format(number, self.grid_point))
        cube = iris.util.squeeze(cube)
        dims = len(cube.shape)
        try:
            cube.coord('i')
        except iris.exceptions.CoordinateNotFoundError:
            cube.add_dim_coord(iris.coords.DimCoord(np.arange(cube.shape[dims - 1]), var_name='i'), dims - 1)
        try:
            cube.coord('j')
        except iris.exceptions.CoordinateNotFoundError:
            cube.add_dim_coord(iris.coords.DimCoord(np.arange(cube.shape[dims - 2]), var_name='j'), dims - 2)
        return cube

    def _load_data(self):
        coords = []
        handler = Utils.open_cdf(self.variable_file.local_file)
        for variable in handler.variables:
            if variable in ('time', 'lev', 'lat', 'lon', 'latitude', 'longitude', 'leadtime', 'time_centered'):
                coords.append(variable)
            if variable == 'time_centered':
                handler.variables[variable].standard_name = ''

        handler.variables[self.variable].coordinates = ' '.join(coords)
        handler.close()

        data = iris.load_cube(self.variable_file.local_file)
        return self._rename_depth(data)

    def _rename_depth(self, data):
        for coord_name in ('model_level_number', 'Vertical T levels', 'lev'):
            if data.coords(coord_name):
                coord = data.coord(coord_name)
                coord.standard_name = 'depth'
                coord.long_name = 'depth'
                break
        return data

    def _fix_file_metadata(self):
        handler = Utils.open_cdf(self.variable_file.local_file)
        var = handler.variables[self.variable]
        coordinates = ''
        has_levels = False
        for dimension in handler.variables.keys():
            if dimension in ['time', 'lev', 'lat', 'latitude', 'lon', 'longitude', 'i', 'j']:
                coordinates += ' {0}'.format(dimension)
            if dimension == 'lev':
                has_levels = True
        var.coordinates = coordinates
        handler.close()
        return has_levels

    def _save_result(self, result, data):
        final_name = '{0}zonal'.format(self.variable)
        temp = TempFile.get()
        handler_source = Utils.open_cdf(self.variable_file.local_file)
        handler_temp = Utils.open_cdf(temp, 'w')
        Utils.copy_variable(handler_source, handler_temp, 'time', True, True)
        handler_temp.createDimension('region', len(result))
        handler_temp.createDimension('region_length', 50)
        handler_temp.createDimension('lev', data.shape[1])
        handler_temp.createDimension('lat', 180)

        var_lat = handler_temp.createVariable('lat', float, 'lat')
        var_lat[...] = np.arange(-90, 90, dtype=np.float32)
        var_lat.units = 'degrees_north'
        var_lat.long_name = 'latitude'
        var_lat.standard_name = 'latitude'

        var_level = handler_temp.createVariable('lev', float, 'lev')
        var_level[...] = data.coord('depth').points
        var_level.units = 'm'
        var_level.axis = 'Z'
        var_level.positive = 'down'
        var_level.long_name = 'ocean depth coordinate'
        var_level.standard_name = 'depth'
        var_region = handler_temp.createVariable('region', 'S1',
                                                 ('region', 'region_length'))
        var = handler_temp.createVariable(
            '{0}zonal'.format(self.variable),
            float,
            ('time', 'lev', 'lat', 'region',),
        )
        var.units = '{0}'.format(data.units)
        for i, basin in enumerate(result):
            var_region[i, ...] = netCDF4.stringtoarr(str(basin), 50)
            var[..., i] = result[basin]
        handler_temp.close()
        self.declared[final_name].set_local_file(temp, diagnostic=self)
